@@ -40,7 +40,6 @@
 ****************************************************************************/
 
 #include "qdeclarativegeomap_p.h"
-#include "qdeclarativegeomapmousearea_p.h"
 #include "error_messages.h"
 
 #include "qdeclarativecirclemapitem_p.h"
@@ -57,9 +56,14 @@
 #include "mapnode_p.h"
 #include <cmath>
 
+#include <QtPositioning/QGeoCoordinate>
+#include <QtPositioning/QGeoCircle>
+#include <QtPositioning/QGeoRectangle>
 #include <QtLocation/QGeoServiceProvider>
 #include <QtLocation/private/qgeomappingmanager_p.h>
+#include "qdoublevector2d_p.h"
 
+#include <QPointF>
 #include <QtQml/QQmlContext>
 #include <QtQml/qqmlinfo.h>
 #include <QModelIndex>
@@ -129,13 +133,6 @@ QT_BEGIN_NAMESPACE
     specially at startup and cannot be replaced or destroyed. Its properties
     can be altered, however, to control its behavior.
 
-    Mouse and touch interaction with Maps and map objects is slightly different
-    to ordinary QML types. In a Map or Map object, you will need to make use
-    of the \l MapMouseArea type instead of the normal Qt Quick MouseArea.
-    MapMouseArea is, in almost all respects, a drop-in replacement for
-    MouseArea, but the documentation for that type should be referred to for
-    further details.
-
     \section2 Performance
 
     Maps are rendered using OpenGL (ES) and the Qt Scene Graph stack, and as
@@ -190,8 +187,7 @@ QDeclarativeGeoMap::QDeclarativeGeoMap(QQuickItem *parent)
         serviceProvider_(0),
         mappingManager_(0),
         zoomLevel_(8.0),
-        bearing_(0.0),
-        tilt_(0.0),
+        center_(51.5073,-0.1277), //London city center
         activeMapType_(0),
         componentCompleted_(false),
         mappingManagerInitialized_(false),
@@ -260,9 +256,9 @@ void QDeclarativeGeoMap::onMapChildrenChanged()
             copyrightsWPtr_ = new QDeclarativeGeoMapCopyrightNotice(this);
             copyrights = copyrightsWPtr_.data();
             connect(map_,
-                    SIGNAL(copyrightsChanged(const QImage &, const QPoint &)),
+                    SIGNAL(copyrightsChanged(QImage,QPoint)),
                     copyrights,
-                    SLOT(copyrightsChanged(const QImage &, const QPoint &)));
+                    SLOT(copyrightsChanged(QImage,QPoint)));
         } else {
             // just re-set its parent.
             copyrights->setParent(this);
@@ -282,8 +278,8 @@ void QDeclarativeGeoMap::pluginReady()
     mappingManager_ = serviceProvider_->mappingManager();
 
     if (!mappingManager_ || serviceProvider_->error() != QGeoServiceProvider::NoError) {
-        QString msg = QCoreApplication::translate(CONTEXT_NAME, PLUGIN_DOESNOT_SUPPORT_MAPPING).arg(serviceProvider_->errorString());
-        qmlInfo(this) << msg;
+        qmlInfo(this) << QStringLiteral("Error: Plugin does not support mapping.\nError message:")
+                      << serviceProvider_->errorString();
         return;
     }
 
@@ -449,7 +445,7 @@ QSGNode *QDeclarativeGeoMap::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeDa
 void QDeclarativeGeoMap::setPlugin(QDeclarativeGeoServiceProvider *plugin)
 {
     if (plugin_) {
-        qmlInfo(this) << QCoreApplication::translate(CONTEXT_NAME, PLUGIN_SET_ONCE);
+        qmlInfo(this) << QStringLiteral("Plugin is a write-once property, and cannot be set again.");
         return;
     }
     plugin_ = plugin;
@@ -473,49 +469,39 @@ void QDeclarativeGeoMap::mappingManagerInitialized()
 
     map_ = mappingManager_->createMap(this);
     gestureArea_->setMap(map_);
-    gestureArea_->zoomLevelLimits(mappingManager_->cameraCapabilities().minimumZoomLevel(),
-                                  mappingManager_->cameraCapabilities().maximumZoomLevel());
+
+    //The zoom level limits are only restricted by the plugins values, if the user has set a more
+    //strict zoom level limit before initialization nothing is done here.
+    if (mappingManager_->cameraCapabilities().minimumZoomLevel() > gestureArea_->minimumZoomLevel())
+        setMinimumZoomLevel(mappingManager_->cameraCapabilities().minimumZoomLevel());
+
+    if (gestureArea_->maximumZoomLevel() < 0
+            || mappingManager_->cameraCapabilities().maximumZoomLevel() < gestureArea_->maximumZoomLevel())
+        setMaximumZoomLevel(mappingManager_->cameraCapabilities().maximumZoomLevel());
 
     map_->setActiveMapType(QGeoMapType());
 
     copyrightsWPtr_ = new QDeclarativeGeoMapCopyrightNotice(this);
     connect(map_,
-            SIGNAL(copyrightsChanged(const QImage &, const QPoint &)),
+            SIGNAL(copyrightsChanged(QImage,QPoint)),
             copyrightsWPtr_.data(),
-            SLOT(copyrightsChanged(const QImage &, const QPoint &)));
+            SLOT(copyrightsChanged(QImage,QPoint)));
 
-    QGeoCameraCapabilities capabilities = mappingManager_->cameraCapabilities();
-    if (zoomLevel_ < capabilities.minimumZoomLevel())
-        setZoomLevel(capabilities.minimumZoomLevel());
-    else if (zoomLevel_ > capabilities.maximumZoomLevel())
-        setZoomLevel(capabilities.maximumZoomLevel());
     connect(map_,
             SIGNAL(updateRequired()),
             this,
             SLOT(update()));
     connect(map_->mapController(),
-            SIGNAL(centerChanged(AnimatableCoordinate)),
+            SIGNAL(centerChanged(QGeoCoordinate)),
             this,
-            SLOT(mapCenterChanged(AnimatableCoordinate)));
-    connect(map_->mapController(),
-            SIGNAL(bearingChanged(qreal)),
-            this,
-            SLOT(mapBearingChanged(qreal)));
-    connect(map_->mapController(),
-            SIGNAL(tiltChanged(qreal)),
-            this,
-            SLOT(mapTiltChanged(qreal)));
+            SIGNAL(centerChanged(QGeoCoordinate)));
     connect(map_->mapController(),
             SIGNAL(zoomChanged(qreal)),
             this,
             SLOT(mapZoomLevelChanged(qreal)));
 
-    AnimatableCoordinate acenter = map_->mapController()->center();
-    acenter.setCoordinate(center_);
-    map_->mapController()->setCenter(acenter);
+    map_->mapController()->setCenter(center_);
     map_->mapController()->setZoom(zoomLevel_);
-    map_->mapController()->setBearing(bearing_);
-    map_->mapController()->setTilt(tilt_);
 
     QList<QGeoMapType> types = mappingManager_->supportedMapTypes();
     for (int i = 0; i < types.size(); ++i) {
@@ -566,6 +552,23 @@ QDeclarativeGeoServiceProvider *QDeclarativeGeoMap::plugin() const
     return plugin_;
 }
 
+/*!
+    \internal
+    Sets the gesture areas minimum zoom level. If the camera capabilities
+    has been set this method honors the boundaries set by it.
+*/
+void QDeclarativeGeoMap::setMinimumZoomLevel(qreal minimumZoomLevel)
+{
+    if (gestureArea_ && minimumZoomLevel >= 0) {
+        if (mappingManagerInitialized_
+                && minimumZoomLevel < mappingManager_->cameraCapabilities().minimumZoomLevel()) {
+            minimumZoomLevel = mappingManager_->cameraCapabilities().minimumZoomLevel();
+        }
+        gestureArea_->setMinimumZoomLevel(minimumZoomLevel);
+        setZoomLevel(qBound<qreal>(minimumZoomLevel, zoomLevel(), maximumZoomLevel()));
+    }
+}
+
 /*!
     \qmlproperty real QtLocation::Map::minimumZoomLevel
 
@@ -577,14 +580,33 @@ QDeclarativeGeoServiceProvider *QDeclarativeGeoMap::plugin() const
 
 qreal QDeclarativeGeoMap::minimumZoomLevel() const
 {
-    if (mappingManager_ && mappingManagerInitialized_)
+    if (gestureArea_->minimumZoomLevel() != -1)
+        return gestureArea_->minimumZoomLevel();
+    else if (mappingManager_ && mappingManagerInitialized_)
         return mappingManager_->cameraCapabilities().minimumZoomLevel();
     else
         return -1.0;
 }
 
 /*!
-\qmlproperty real QtLocation::Map::maximumZoomLevel
+    \internal
+    Sets the gesture areas maximum zoom level. If the camera capabilities
+    has been set this method honors the boundaries set by it.
+*/
+void QDeclarativeGeoMap::setMaximumZoomLevel(qreal maximumZoomLevel)
+{
+    if (gestureArea_ && maximumZoomLevel >= 0) {
+        if (mappingManagerInitialized_
+                && maximumZoomLevel > mappingManager_->cameraCapabilities().maximumZoomLevel()) {
+            maximumZoomLevel = mappingManager_->cameraCapabilities().maximumZoomLevel();
+        }
+        gestureArea_->setMaximumZoomLevel(maximumZoomLevel);
+        setZoomLevel(qBound<qreal>(minimumZoomLevel(), zoomLevel(), maximumZoomLevel));
+    }
+}
+
+/*!
+    \qmlproperty real QtLocation::Map::maximumZoomLevel
 
     This property holds the maximum valid zoom level for the map.
 
@@ -594,87 +616,14 @@ qreal QDeclarativeGeoMap::minimumZoomLevel() const
 
 qreal QDeclarativeGeoMap::maximumZoomLevel() const
 {
-    if (mappingManager_ && mappingManagerInitialized_)
+    if (gestureArea_->maximumZoomLevel() != -1)
+        return gestureArea_->maximumZoomLevel();
+    else if (mappingManager_ && mappingManagerInitialized_)
         return mappingManager_->cameraCapabilities().maximumZoomLevel();
     else
         return -1.0;
 }
 
-/*!
-    \internal
-*/
-void QDeclarativeGeoMap::setBearing(qreal bearing)
-{
-    if (bearing_ == bearing)
-        return;
-    bool clockwise = (bearing >= 0);
-    qreal fractions = bearing - int(bearing);
-    bearing = (int(qAbs(bearing))) % 359;
-    if (!clockwise)
-        bearing = (-1.0 * bearing) + 360;
-    bearing_ = bearing + fractions;
-    if (mappingManagerInitialized_)
-        map_->mapController()->setBearing(bearing_);
-    emit bearingChanged(bearing_);
-}
-
-/*!
-    \qmlproperty real Map::bearing
-
-    This property holds the current bearing (starting from 0 and increasing
-    clockwise to 359,9 degrees) pointing up.
-
-    For example setting bearing to 10 will set bearing 10 to point up, which
-    visually looks like rotating the map counter-clockwise.
-
-    You can also assign negative values, which will internally get
-    translated into positive bearing (for example -10 equals 350). This is primarily for
-    convenience (for example you can decrement bearing without worrying about it).
-    Assigning values greater than abs(360) will be mod'd (for example 365 will result
-    in 5).
-
-    The default value is 0 corresponding North pointing up.
-*/
-
-qreal QDeclarativeGeoMap::bearing() const
-{
-    if (mappingManagerInitialized_) {
-        if (map_->mapController()->bearing() >= 0)
-            return map_->mapController()->bearing();
-        else
-            return map_->mapController()->bearing() + 360;
-    } else {
-        return bearing_;
-    }
-}
-
-/*!
-    \qmlproperty real Map::tilt
-
-    This property holds the current tilt (starting from 0 and increasing to 85 degrees).
-
-    The tilt gives the map a 2.5D feel. Certain map objects may be rendered
-    in 3D if the tilt is different from 0.
-
-    The default value is 0 corresponding to no tilt.
-*/
-qreal QDeclarativeGeoMap::tilt() const
-{
-    if (!mappingManagerInitialized_)
-        return tilt_;
-    return map_->mapController()->tilt();
-}
-
-void QDeclarativeGeoMap::setTilt(qreal tilt)
-{
-    if (tilt_ == tilt || tilt > 85.0 || tilt < 0)
-        return;
-    tilt_ = tilt;
-    if (mappingManagerInitialized_)
-        map_->mapController()->setTilt(tilt);
-    emit tiltChanged(tilt);
-}
-
 /*!
     \qmlproperty real QtLocation::Map::zoomLevel
 
@@ -688,13 +637,9 @@ void QDeclarativeGeoMap::setZoomLevel(qreal zoomLevel)
     if (zoomLevel_ == zoomLevel || zoomLevel < 0)
         return;
 
-    if (mappingManagerInitialized_) {
-        QGeoCameraCapabilities capabilities = mappingManager_->cameraCapabilities();
-            if ((zoomLevel < capabilities.minimumZoomLevel() ||
-                    zoomLevel > capabilities.maximumZoomLevel())) {
-                return;
-            }
-    }
+    if ((zoomLevel < minimumZoomLevel()
+         || (maximumZoomLevel() >= 0 && zoomLevel > maximumZoomLevel())))
+        return;
 
     zoomLevel_ = zoomLevel;
     if (mappingManagerInitialized_)
@@ -711,24 +656,25 @@ qreal QDeclarativeGeoMap::zoomLevel() const
 }
 
 /*!
-\qmlproperty coordinate QtLocation::Map::center
+    \qmlproperty coordinate QtLocation::Map::center
 
     This property holds the coordinate which occupies the center of the
-    mapping viewport.
+    mapping viewport. Invalid center coordinates are ignored.
 
     The default value is an arbitrary valid coordinate.
 */
 void QDeclarativeGeoMap::setCenter(const QGeoCoordinate &center)
 {
-    if (center == center_)
+    if (!mappingManagerInitialized_ && center == center_)
+        return;
+
+    if (!center.isValid())
         return;
 
     center_ = center;
 
     if (center_.isValid() && mappingManagerInitialized_) {
-        AnimatableCoordinate acoord = map_->mapController()->center();
-        acoord.setCoordinate(center_);
-        map_->mapController()->setCenter(acoord);
+        map_->mapController()->setCenter(center_);
         update();
     } else {
         emit centerChanged(center_);
@@ -738,7 +684,7 @@ void QDeclarativeGeoMap::setCenter(const QGeoCoordinate &center)
 QGeoCoordinate QDeclarativeGeoMap::center() const
 {
     if (mappingManagerInitialized_)
-        return map_->mapController()->center().coordinate();
+        return map_->mapController()->center();
     else
         return center_;
 }
@@ -754,36 +700,6 @@ void QDeclarativeGeoMap::mapZoomLevelChanged(qreal zoom)
     emit zoomLevelChanged(zoomLevel_);
 }
 
-/*!
-    \internal
-*/
-void QDeclarativeGeoMap::mapTiltChanged(qreal tilt)
-{
-    if (tilt == zoomLevel_)
-        return;
-    tilt_ = tilt;
-    emit tiltChanged(tilt_);
-}
-
-/*!
-    \internal
-*/
-void QDeclarativeGeoMap::mapBearingChanged(qreal bearing)
-{
-    if (bearing == bearing_)
-        return;
-    bearing_ = bearing;
-    emit bearingChanged(bearing_);
-}
-
-/*!
-    \internal
-*/
-void QDeclarativeGeoMap::mapCenterChanged(AnimatableCoordinate center)
-{
-    emit centerChanged(center.coordinate());
-}
-
 /*!
     \qmlproperty list<MapType> QtLocation::Map::supportedMapTypes
 
@@ -809,7 +725,7 @@ QQmlListProperty<QDeclarativeGeoMapType> QDeclarativeGeoMap::supportedMapTypes()
 QGeoCoordinate QDeclarativeGeoMap::toCoordinate(const QPointF &screenPosition) const
 {
     if (map_)
-        return map_->screenPositionToCoordinate(screenPosition);
+        return map_->screenPositionToCoordinate(QDoubleVector2D(screenPosition));
     else
         return QGeoCoordinate();
 }
@@ -827,7 +743,7 @@ QGeoCoordinate QDeclarativeGeoMap::toCoordinate(const QPointF &screenPosition) c
 QPointF QDeclarativeGeoMap::toScreenPosition(const QGeoCoordinate &coordinate) const
 {
     if (map_)
-        return map_->coordinateToScreenPosition(coordinate);
+        return map_->coordinateToScreenPosition(coordinate).toPointF();
     else
         return QPointF(qQNaN(), qQNaN());
 }
@@ -841,7 +757,7 @@ QPointF QDeclarativeGeoMap::toScreenPosition(const QGeoCoordinate &coordinate) c
     Positive values for \a dx move the map right, negative values left.
     Positive values for \a dy move the map down, negative values up.
 
-    During panning the \l center, \l tilt, \l bearing, and \l zoomLevel may change.
+    During panning the \l center, and \l zoomLevel may change.
 */
 void QDeclarativeGeoMap::pan(int dx, int dy)
 {
@@ -884,6 +800,7 @@ void QDeclarativeGeoMap::wheelEvent(QWheelEvent *event)
 {
     QLOC_TRACE0;
     event->accept();
+    gestureArea_->wheelEvent(event);
     emit wheelAngleChanged(event->angleDelta());
 }
 
@@ -898,11 +815,25 @@ bool QDeclarativeGeoMap::childMouseEventFilter(QQuickItem *item, QEvent *event)
     case QEvent::MouseButtonPress:
     case QEvent::MouseButtonRelease:
     case QEvent::MouseMove:
+        if (item->keepMouseGrab())
+            return false;
+        if (!gestureArea_->filterMapChildMouseEvent(static_cast<QMouseEvent *>(event)))
+            return false;
+        grabMouse();
+        return true;
+    case QEvent::UngrabMouse:
         return gestureArea_->filterMapChildMouseEvent(static_cast<QMouseEvent *>(event));
     case QEvent::TouchBegin:
     case QEvent::TouchUpdate:
     case QEvent::TouchEnd:
-        return gestureArea_->filterMapChildTouchEvent(static_cast<QTouchEvent *>(event));
+        if (item->keepMouseGrab())
+            return false;
+        if (!gestureArea_->filterMapChildTouchEvent(static_cast<QTouchEvent *>(event)))
+            return false;
+        grabMouse();
+        return true;
+    case QEvent::Wheel:
+        return gestureArea_->wheelEvent(static_cast<QWheelEvent *>(event));
     default:
         return false;
     }
@@ -1045,6 +976,85 @@ void QDeclarativeGeoMap::geometryChanged(const QRectF &newGeometry, const QRectF
     QQuickItem::geometryChanged(newGeometry, oldGeometry);
 }
 
+/*!
+    \qmlmethod QtLocation::Map::fitViewportToGeoShape(QGeoShape shape)
+
+    Fits the current viewport to the boundary of the shape. The camera is positioned
+    in the center of the shape, and at the largest integral zoom level possible which
+    allows the whole shape to be visible on screen
+
+*/
+void QDeclarativeGeoMap::fitViewportToGeoShape(const QVariant &variantShape)
+{
+    if (!map_ || !mappingManagerInitialized_)
+        return;
+
+    QGeoShape shape;
+
+    if (variantShape.userType() == qMetaTypeId<QGeoRectangle>())
+        shape = variantShape.value<QGeoRectangle>();
+    else if (variantShape.userType() == qMetaTypeId<QGeoCircle>())
+        shape = variantShape.value<QGeoCircle>();
+    else if (variantShape.userType() == qMetaTypeId<QGeoShape>())
+        shape = variantShape.value<QGeoShape>();
+
+    if (!shape.isValid())
+        return;
+
+    double bboxWidth;
+    double bboxHeight;
+    QGeoCoordinate centerCoordinate;
+
+    switch (shape.type()) {
+    case QGeoShape::RectangleType:
+    {
+        QGeoRectangle rect = shape;
+        QDoubleVector2D topLeftPoint = map_->coordinateToScreenPosition(rect.topLeft(), false);
+        QDoubleVector2D botRightPoint = map_->coordinateToScreenPosition(rect.bottomRight(), false);
+        bboxWidth = qAbs(topLeftPoint.x() - botRightPoint.x());
+        bboxHeight = qAbs(topLeftPoint.y() - botRightPoint.y());
+        centerCoordinate = rect.center();
+        break;
+    }
+    case QGeoShape::CircleType:
+    {
+        QGeoCircle circle = shape;
+        centerCoordinate = circle.center();
+        QGeoCoordinate edge = centerCoordinate.atDistanceAndAzimuth(circle.radius(), 90);
+        QDoubleVector2D centerPoint = map_->coordinateToScreenPosition(centerCoordinate, false);
+        QDoubleVector2D edgePoint = map_->coordinateToScreenPosition(edge, false);
+        bboxWidth = qAbs(centerPoint.x() - edgePoint.x()) * 2;
+        bboxHeight = bboxWidth;
+        break;
+    }
+    case QGeoShape::UnknownType:
+        //Fallthrough to default
+    default:
+        return;
+    }
+
+    // position camera to the center of bounding box
+    setProperty("center", QVariant::fromValue(centerCoordinate));
+
+    //If the shape is empty we just change centerposition, not zoom
+    if (bboxHeight == 0 && bboxWidth == 0)
+        return;
+
+    // adjust zoom
+    double bboxWidthRatio = bboxWidth / (bboxWidth + bboxHeight);
+    double mapWidthRatio = width() / (width() + height());
+    double zoomRatio;
+
+    if (bboxWidthRatio > mapWidthRatio)
+        zoomRatio = bboxWidth / width();
+    else
+        zoomRatio = bboxHeight / height();
+
+    qreal newZoom = log10(zoomRatio) / log10(0.5);
+    newZoom = floor(qMax(minimumZoomLevel(), (map_->mapController()->zoom() + newZoom)));
+    setProperty("zoomLevel", QVariant::fromValue(newZoom));
+}
+
 /*!
     \qmlmethod QtLocation::Map::fitViewportToMapItems()
 
@@ -1068,18 +1078,18 @@ void QDeclarativeGeoMap::fitViewportToMapItemsRefine(bool refine)
     if (mapItems_.size() == 0)
         return;
 
-    qreal minX = 0;
-    qreal maxX = 0;
-    qreal minY = 0;
-    qreal maxY = 0;
-    qreal topLeftX = 0;
-    qreal topLeftY = 0;
-    qreal bottomRightX = 0;
-    qreal bottomRightY = 0;
+    double minX = 0;
+    double maxX = 0;
+    double minY = 0;
+    double maxY = 0;
+    double topLeftX = 0;
+    double topLeftY = 0;
+    double bottomRightX = 0;
+    double bottomRightY = 0;
 
     // find bounds of all map items
     QGeoCoordinate geoCenter;
-    QPointF centerPt;
+    QDoubleVector2D centerPt;
     int itemCount = 0;
     for (int i = 0; i < mapItems_.count(); ++i) {
         if (!mapItems_.at(i))
@@ -1131,17 +1141,15 @@ void QDeclarativeGeoMap::fitViewportToMapItemsRefine(bool refine)
     if (itemCount == 0)
         return;
 
-    qreal bboxWidth = maxX - minX;
-    qreal bboxHeight = maxY - minY;
-    qreal bboxCenterX = minX + (bboxWidth / 2.0);
-    qreal bboxCenterY = minY + (bboxHeight / 2.0);
+    double bboxWidth = maxX - minX;
+    double bboxHeight = maxY - minY;
+    double bboxCenterX = minX + (bboxWidth / 2.0);
+    double bboxCenterY = minY + (bboxHeight / 2.0);
 
     // position camera to the center of bounding box
     QGeoCoordinate coordinate;
-    coordinate = map_->screenPositionToCoordinate(QPointF(bboxCenterX, bboxCenterY), false);
-    AnimatableCoordinate acenter = map_->mapController()->center();
-    acenter.setCoordinate(coordinate);
-    setCenter(coordinate);
+    coordinate = map_->screenPositionToCoordinate(QDoubleVector2D(bboxCenterX, bboxCenterY), false);
+    setProperty("center", QVariant::fromValue(coordinate));
 
     // adjust zoom
     double bboxWidthRatio = bboxWidth / (bboxWidth + bboxHeight);
@@ -1155,7 +1163,7 @@ void QDeclarativeGeoMap::fitViewportToMapItemsRefine(bool refine)
 
     qreal newZoom = log10(zoomRatio) / log10(0.5);
     newZoom = floor(qMax(minimumZoomLevel(), (map_->mapController()->zoom() + newZoom)));
-    setZoomLevel(newZoom);
+    setProperty("zoomLevel", QVariant::fromValue(newZoom));
 
     // as map quick items retain the same screen size after the camera zooms in/out
     // we refine the viewport again to achieve better results
